import sys
import javalang as JL
import xml.etree.ElementTree as ET
import re
import random
from itertools import chain
import glob
from z3 import *
from z3printer import *

# get the directory where the java files are in javaSourceDir
javaSourceDir = sys.argv[1]

# get all java file names in the directory and keep one in javaFile
javaFileNames = [] 
for fileName in os.listdir(javaSourceDir):
    if fileName.endswith(".java"):
        javaFileNames.append(fileName)
javaFileName = javaFileNames[0]
javaFile = open(javaSourceDir + javaFileName)

# get all methods from java file and keep one in method
javaFileContent = javaFile.read()
javaTree = JL.parse.parse(javaFileContent)
methods = []
inputNames = {}
outputName = None
for path, node in javaTree:
    if isinstance(node, JL.tree.MethodDeclaration):
        methods.append(node)
method = methods[0]

# get method, inputs, and output names in methodName, inputNames, and outputName
methodName = method.name
for param in method.parameters:
    inputNames[param.name] = {}
for path, node in method:
    if isinstance(node, JL.tree.ReturnStatement):
        outputName = node.expression.member

# get result of symbolic execution in symbExecFile (waiting for KeY API improvement, generated from Eclipse SED Plug-in)
symbExecFile = open(javaSourceDir + javaFileName.split(".")[0] + ".xml")

# get path conditions and states in pathConditions and states
symbExecContent = symbExecFile.read()
symbExecRoot = ET.fromstring(symbExecContent)
pathConditions = []
states = []
for child in symbExecRoot[0]:
    if (child.attrib)["terminationKind"] == "NORMAL":
        pathConditions.append((child.attrib)["pathCondition"])
        states.append((child.attrib)["state"])

# get variables from path conditions and states in symbVariables
symbVariablesSplitted = []
for pc in chain(pathConditions, states):
    symbVariablesSplitted.append(re.sub(r" +\d+ +", " ", re.sub(r"[==|<=|<|>|>=|&|+|*|-|!=]+", " ", pc)))
symbVariables = ""
for e in symbVariablesSplitted:
    symbVariables = symbVariables + e
symbVariables = symbVariables.split(" ")
symbVariables = list(set(symbVariables))
if "" in symbVariables:
    symbVariables.remove("")
if "-" in symbVariables:
    symbVariables.remove("-")

# get z3 types in z3Variables
z3Variables = {}
for var in symbVariables:
    z3Variables[var] = Int(var)

# get global preconditions in preconditions
precondition = (symbExecRoot[0].attrib)["contractText"]
preconditions = precondition.split("&")

# get global preconditions string in preconditionsString
preconditionsString = "And( "
for pc in preconditions:
    preconditionsString = preconditionsString + pc + " , "
preconditionsString = preconditionsString[:-2] + ")"
for var in symbVariables:
    preconditionsString = preconditionsString.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)

# get path conditions strings in pathConditionsStrings
pathConditionsSplitted = []
for pc in pathConditions:
    pathConditionsSplitted.append(pc.split("&"))
pathConditionsStrings = []
for pc in pathConditionsSplitted:
    pathConditionString = "And( "
    for atom in pc:
        pathConditionString = pathConditionString + atom + " , "
    pathConditionString = pathConditionString[:-2] + ")"
    pathConditionsStrings.append(pathConditionString)

# get states strings in statesStrings
statesSplitted = []
for st in states:
    statesSplitted.append(st.split("&"))
statesStrings = []
for st in statesSplitted:
    stateStrings = "And( "
    for atom in st:
        stateStrings = stateStrings + atom + " , "
    stateStrings = stateStrings[:-2] + ")"
    statesStrings.append(stateStrings)

# get leaf string in leafString
leafString = "Or( "
for l in range(0, len(pathConditionsStrings)):
    leafString = leafString + " And( " + pathConditionsStrings[l] + " , " + statesStrings[l] + " ), "
leafString = leafString[:-2] + ")"
for var in symbVariables:
    leafString = leafString.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)

# get z3 symbolic characterization in z3SymbolicCharacterization
symbolicCharacterizationString = "And(" + preconditionsString + ", " + leafString + ")"
z3SymbolicCharacterization = eval(symbolicCharacterizationString)

# get input limits in inputNames
for atom in preconditions:
    atom = atom + "_symb"
    r_lt = re.compile(' *(\w+) *< *(\w+) *')
    r_lteq = re.compile(' *(\w+) *<= *(\w+) *')
    r_eq = re.compile(' *(\w+) *== *(\w+) *')
    r_gt = re.compile(' *(\w+) *> *(\w+) *')
    r_gteq = re.compile(' *(\w+) *>= *(\w+) *')
    m_lt = r_lt.search(atom)
    m_lteq = r_lteq.search(atom)
    m_eq = r_eq.search(atom)
    m_gt = r_gt.search(atom)
    m_gteq = r_gteq.search(atom)
    for var in inputNames:
        if m_lt:
            if (var + "_symb") == m_lt.group(1):
                inputNames[var]["max"] = int(m_lt.group(2)) - 1
            elif (var + "_symb") == m_lt.group(2):
                inputNames[var]["min"] = int(m_lt.group(1)) + 1
        if m_lteq:
            if (var + "_symb") == m_lteq.group(1):
                inputNames[var]["max"] = int(m_lteq.group(2))
            elif (var + "_symb") == m_lteq.group(2):
                inputNames[var]["min"] = int(m_lteq.group(1))
        if m_eq:
            if (var + "_symb") == m_eq.group(1):
                inputNames[var]["min"] = int(m_eq.group(2))
                inputNames[var]["max"] = int(m_eq.group(2))
            elif (var + "_symb") == m_eq.group(2):
                inputNames[var]["min"] = int(m_eq.group(1))
                inputNames[var]["max"] = int(m_eq.group(1))
        if m_gt:
            if (var + "_symb") == m_gt.group(1):
                inputNames[var]["min"] = int(m_gt.group(2)) + 1
            elif (var + "_symb") == m_gt.group(2):
                inputNames[var]["max"] = int(m_gt.group(1)) - 1
        if m_gteq:
            if (var + "_symb") == m_gteq.group(1):
                inputNames[var]["min"] = int(m_gteq.group(2))
            elif (var + "_symb") == m_gteq.group(2):
                inputNames[var]["max"] = int(m_gteq.group(1))

# get input domain in inputDomain
inputDomain = {}
for i in inputNames:
    inputDomain[i] = range(inputNames[i]["min"], inputNames[i]["max"]+1)

# load solver and z3SymbolicCharacterization
s = Solver()
s.add(z3SymbolicCharacterization)

# compute kernel function in z3k
if (len(inputNames) == 2):
    z3k = {}
    for inp1 in inputDomain:
        inp2CursorDomain = inputDomain.keys()
        inp2CursorDomain.remove(inp1)
        z3kinp1 = {}
        for val1 in inputDomain[inp1]:
            val1CursorDomain = list(inputDomain[inp1])
            val1CursorDomain.remove(val1)
            tempEquivClass = [val1]
            for val1Cursor in val1CursorDomain:
                equalityNumber = 0
                for inp2Cursor in inp2CursorDomain:
                    for val2Cursor in inputDomain[inp2Cursor]:
                        s.push()
                        caseString1 = " " + inp1 + " == " + str(val1) + " "
                        caseString2 = " " + inp2Cursor + " == " + str(val2Cursor) + " "
                        for var in inputDomain:
                            caseString1 = caseString1.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
                            caseString2 = caseString2.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
                        z3case1 = eval(caseString1)
                        z3case2 = eval(caseString2)
                        s.add(z3case1)
                        s.add(z3case2)
                        if s.check() == sat:
                            o1 = s.model()[z3Variables[outputName]]
                        s.pop()
                        s.push()
                        caseString1 = " " + inp1 + " == " + str(val1Cursor) + " "
                        caseString2 = " " + inp2Cursor + " == " + str(val2Cursor) + " "
                        for var in inputDomain:
                            caseString1 = caseString1.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
                            caseString2 = caseString2.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
                        z3case1 = eval(caseString1)
                        z3case2 = eval(caseString2)
                        s.add(z3case1)
                        s.add(z3case2)
                        if s.check() == sat:
                            o2 = s.model()[z3Variables[outputName]]
                        s.pop()
                        if o1.sexpr() == o2.sexpr():
                            equalityNumber = equalityNumber + 1
                if equalityNumber == len(inputDomain[inp2Cursor]):
                    tempEquivClass.append(val1Cursor)
            z3kinp1[val1] = tempEquivClass
        z3k[inp1] = z3kinp1
if (len(inputNames) == 1):
    z3k = {}
    for inp in inputDomain:
        z3kinp = {}
        for val in inputDomain[inp]:
            tempEquivClass = []
            s.push()
            caseString = " " + inp + " == " + str(val) + " "
            for var in inputDomain:
                caseString = caseString.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
            z3case = eval(caseString)
            s.add(z3case)
            if s.check() == sat:
                o1 = s.model()[z3Variables[outputName]]
            s.pop()
            for valCursor in inputDomain[inp]:
                s.push()
                caseString = " " + inp + " == " + str(valCursor) + " "
                for var in inputDomain:
                    caseString = caseString.replace(" " + var + " ", """ z3Variables[\"""" + var + """\"] """)
                z3case = eval(caseString)
                s.add(z3case)
                #print s.check()
                if s.check() == sat:
                    o2 = s.model()[z3Variables[outputName]]
                s.pop()
                if o1.sexpr() == o2.sexpr():
                    tempEquivClass.append(valCursor)
            z3kinp[val] = tempEquivClass
        z3k[inp] = z3kinp

# get section function in z3r
z3r = {}
for inp in z3k:
    z3rinp = {}
    for val in z3k[inp]:
        z3rinp[val] = min(z3k[inp][val])
    z3r[inp] = z3rinp

# get postprocessing instructions (by heuristics) of data in minimiserTreeList
minimiserTreeTuple = {}
for inp in z3k:
    minimiserTreeTuple[inp] = sorted(z3k[inp].items())
    for pair in minimiserTreeTuple[inp]:
        pair[1].sort()
minimiserTreeList = {}
for inp in minimiserTreeTuple:
    minimiserTreeList[inp] = []
    for pair in minimiserTreeTuple[inp]:
        minimiserTreeList[inp].append([[pair[0]], pair[1]])
# apply representative non minimisable: input = input
for inp in minimiserTreeList:
    i = 0
    length = len(minimiserTreeList[inp])
    while i < length:
        if len(minimiserTreeList[inp][i][1]) == 1:
            minimiserTreeList[inp][i][1] = [inp]
        i = i+1
# apply representative by z3r
for inp in minimiserTreeList:
    i = 0
    length = len(minimiserTreeList[inp])
    while i < length:
        if len(minimiserTreeList[inp][i][1]) != 1:
            toRepr = minimiserTreeList[inp][i][0][0]
            minimiserTreeList[inp][i][1] = [z3r[inp][toRepr]]
        i = i+1
# contiguous values: [0, 1] = [0, 1]
for inp in minimiserTreeList:
    i = 0
    length = len(minimiserTreeList[inp])
    while i < length:
        if i < length-1:
            if minimiserTreeList[inp][i][1] == minimiserTreeList[inp][i+1][1]:
                minimiserTreeList[inp][i][0].extend(minimiserTreeList[inp][i+1][0])
                del minimiserTreeList[inp][i+1]
                i = i-1
                length = len(minimiserTreeList[inp])
        i = i+1

# generate ClassMin.java minimiser in minimiserStrings
minimiserStrings = {}
for inp in inputNames:
    minimiserStrings[inp] = "public class " + javaFileName.split(".")[0] + "Min_" + inp + " {\n    public int minimise_" + methodName + "(int " + inp + ") {\n        int repr_" + inp + " = " + str(minimiserTreeList[inp][0][1][0]) + ";\n"
    if len(minimiserTreeList[inp]) == 2:
        minimiserStrings[inp] = minimiserStrings[inp] + "        if (" + inp + " >= " + str(minimiserTreeList[inp][1][0][0]) + ") { repr_" + inp + " = " + str(minimiserTreeList[inp][1][1][0]) + "; }\n"
    else:
        i = 1
        length = len(minimiserTreeList[inp])
        while i < length-1:
            minimiserStrings[inp] = minimiserStrings[inp] + "        if (" + inp + " >= " + str(minimiserTreeList[inp][i][0][0]) + " && " + inp + " <= " + str(minimiserTreeList[inp][i][0][-1]) + ") { repr_" + inp + " = " + str(minimiserTreeList[inp][i][1][0]) + "; }\n"
            i = i+1
        minimiserStrings[inp] = minimiserStrings[inp] + "        if (" + inp + " >= " + str(minimiserTreeList[inp][i][0][0]) + ") { repr_" + inp + " = " + str(minimiserTreeList[inp][i][1][0]) + "; }\n"    
    minimiserStrings[inp] = minimiserStrings[inp] + "        return repr_" + inp + "; } }"

# put minimiser in Minimiser directory
currentWorkDirectory = os.getcwd()
if os.path.isdir(currentWorkDirectory + "/" + javaSourceDir + "Minimiser"):
    files = glob.glob(currentWorkDirectory + "/" + javaSourceDir + "Minimiser/*")
    for f in files:
        os.remove(f)
else:
    os.makedirs(javaSourceDir + "Minimiser")
os.chdir(currentWorkDirectory + "/" + javaSourceDir + "Minimiser/")
for inp in inputNames:
    minimiserFile = open(javaFileName.split(".")[0] + "Min_" + inp + ".java", "w+")
    minimiserFile.write(minimiserStrings[inp])
    minimiserFile.close()
os.chdir(currentWorkDirectory)
     
# get number of equivalence classes in countEqClasses
countEqClasses = {}
for inp in inputNames:
    countingSet = set()
    for val in z3k[inp]:
        eq = frozenset(z3k[inp][val])
        countingSet.add(eq)
    countEqClasses[inp] = len(countingSet)
   
# print content of variables
print "------------------------------------------"
print "methodName: " + methodName
print "inputNames: " + str(inputNames)
print "outputName: " + outputName
print "z3Variables: " + str(z3Variables)
print "inputDomain: " + str(inputDomain)

# print (equivalence classes) and representatives
print "------------------------------------------"
#print "The equivalences classes are: " + str(z3k)
print "The representatives chosen are: " + str(minimiserTreeList)

# print diagnosis
print "------------------------------------------"
for inp in inputNames:
    print "The cardinal of the domain of the input " + str(inp) + " was " + str(len(inputDomain[inp])) + ", it has been minimised to " + str(countEqClasses[inp]) + " class(es)."

# print end messages
print "------------------------------------------"
print "The minimiser has been correctly generated."
print "It is composed of " + str(len(inputNames)) + " java files."
print "See in " + currentWorkDirectory + "/" + javaSourceDir + "Minimiser/"
print "------------------------------------------"